\begin{tabular}{ {{ alignment }} }
{% if caption -%}
% The caption text
\multicolumn{ {{ column_names|length }} }{c}{ {{ caption }} } \\[0.2cm]
%\vspace*{0.2cm}
{% endif -%}
% -- table header --
\toprule
{% for name in column_names -%}
{{ name }} {% if loop.last %}\\{% else %}&{% endif %}
{% endfor %}
\midrule
% -- table content --
{{ content }}
\bottomrule
\end{tabular}

